
/* PPI AMD related command functions */

#include <linux/module.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <linux/mtd/nor.h> 
/* erase and write commnads interval counts from h/w engineer */
static unsigned long ppi_write_wait_count = 0x500;
static unsigned long ppi_erase_wait_count = 0x1050000;

int ppi_amd_init(void)
{
	return 0;
}

int ppi_amd_reset(void)
{
	writel(0xf0, PIO_DATA_REG);
	
	return 0;
}

int ppi_amd_read_id(u8 *id)
{
	writel(0xAAA, NOR_ADDR_REG);
	writel(0xAA,  PIO_DATA_REG);
	writel(0x555, NOR_ADDR_REG);
	writel(0x55,  PIO_DATA_REG);
	writel(0xAAA, NOR_ADDR_REG);
	writel(0x90,  PIO_DATA_REG);
	writel(0x02,  NOR_ADDR_REG);
	*id=readb(PIO_DATA_REG);
	ppi_amd_reset();

	return 0;
}

int ppi_amd_cmd_erase(unsigned long addr, int mode)
{
	int rc = 0;
	unsigned char stat;

	switch ( mode ) {
	case DMA_START:
		writel(ppi_erase_wait_count, PPI_PRO_CNT_REG);

		/* write-protected disable, auto-mode, 6 sets of commands */
		writel(0x36, COM_REG); 
	
		writel(0xAAA, ADDR_CFG1);
		writel(0xAA,  DATA_CFG1);
		writel(0x555, ADDR_CFG2);
		writel(0x55,  DATA_CFG2);
		writel(0xAAA, ADDR_CFG3);
		writel(0x80,  DATA_CFG3);
		writel(0xAAA, ADDR_CFG4);
		writel(0xAA,  DATA_CFG4);
		writel(0x555, ADDR_CFG5);
		writel(0x55,  DATA_CFG5);
	
		writel(addr, NOR_ADDR_REG);
		writeb(0x30, PIO_DATA_REG);
		break;
	case DMA_END:
		stat = readb(PIO_DATA_REG)&0x80; /* DATA_POLLING 0x80 */
		if ( stat == 0 )
			rc = -1;
		else
			rc = 0;
		writel(0x00, COM_REG);
		break;
	}
	
	return rc;
}

int ppi_amd_cmd_write(unsigned long addr, int size, int mode)
{
	int rc = 0;
	
	switch ( mode ) {
	case DMA_START:
		writel(ppi_write_wait_count, PPI_PRO_CNT_REG);

		/* write-protected disable, auto-mode, 4 sets of commands */
		writel(0x34, COM_REG); 
	
		writel(0xAAA, ADDR_CFG1);
		writel(0xAA,  DATA_CFG1);
		writel(0x555, ADDR_CFG2);
		writel(0x55,  DATA_CFG2);
		writel(0xAAA, ADDR_CFG3);
		writel(0xA0,  DATA_CFG3);
	
		writel(addr, NOR_ADDR_REG);
		writel(0x80000000|size, DMA_S_REG); 
		break;
	case DMA_END:
		writel(0x00, COM_REG);
		break;
	}
	                                         
	return rc;
}

int ppi_amd_cmd_read(unsigned long addr, int size, int mode)
{
	int rc = 0;
	
	switch ( mode ) {
	case DMA_START:
		writel(addr, NOR_ADDR_REG);
		writel(0x40000000|size, DMA_S_REG);
		break;
	case DMA_END:
		/* nothing to do */
		break;
	}
	
	return rc;
}

int ppi_amd_pio_read(unsigned long addr, char *buf, int size)
{
	char *p;
	int val, remains, count, rc;
	
	rc = 0;
	remains = size;
	p = buf;

	writel(0x30, COM_REG);
	while ( remains > 0 ) {
		writel(addr, NOR_ADDR_REG);
		if ( remains >= 4 ) {
			writel(readl(PIO_DATA_REG), p);
			count = 4;
		} else {
			val = readl(PIO_DATA_REG);
			memcpy(p, (char *)&val, remains);
			count = remains;
		}
		addr += count;
		p += count;
		remains -= count;
	}
	writel(0x00, COM_REG);
	
	return rc;
}

int ppi_amd_pio_write(unsigned long addr, char *buf, int size)
{
	char *p;
	int val, remains, count, rc, ic;

	writel(ppi_write_wait_count, PPI_PRO_CNT_REG);
	writel(0x34,  COM_REG);
	writel(0xAAA, ADDR_CFG1);
	writel(0xAA,  DATA_CFG1);
	writel(0x555, ADDR_CFG2);
	writel(0x55,  DATA_CFG2);
	writel(0xAAA, ADDR_CFG3);
	writel(0xA0,  DATA_CFG3);

	rc = 0;
	remains = size;
	p = buf;
	
	
	if (( addr & 0x01 )||((unsigned long)p & 0x01)) {
		while ( remains > 0 ) {
			writel(addr, NOR_ADDR_REG);
			writeb(readb(p),PIO_DATA_REG);	
			rc = pl_nor_polling_intr(NOR_CMD_INTR_MASK, 0);
			//printk("remains=%d , src 0x%p , val=0x%x, to target 0x%x\n",remains,p,readb(p),addr);
			if ( rc == 0 ){
				printk("pl_nor_polling_intr, rc=0\n");
				remains=0;
				break;
			}
			addr++;
			p++;
			remains--;
		}			
	}
	
	while ( remains > 0 ) {
		writel(addr, NOR_ADDR_REG);
		if ( remains >= 4 ) {
			writel(readl(p), PIO_DATA_REG);
			count = 4;
		} else {
			memset((char *)&val, 0xff, 4);
			memcpy((char *)&val, p, remains);
			writel(val,PIO_DATA_REG);				
			count = remains;
		}
		rc = pl_nor_polling_intr(NOR_CMD_INTR_MASK, 0);
		if ( rc == 0 )
			break;
		addr += count;
		p += count;
		remains -= count;
	}
	writel(0x00, COM_REG);
	readb(INTR_S_REG);	//read clean INTR_S_REG
	return rc;
}


EXPORT_SYMBOL(ppi_amd_init);
EXPORT_SYMBOL(ppi_amd_reset);
EXPORT_SYMBOL(ppi_amd_read_id);
EXPORT_SYMBOL(ppi_amd_cmd_erase);
EXPORT_SYMBOL(ppi_amd_cmd_write);
EXPORT_SYMBOL(ppi_amd_cmd_read);
EXPORT_SYMBOL(ppi_amd_pio_read);
EXPORT_SYMBOL(ppi_amd_pio_write);
